/*
**    TP 5ETI CPE Lyon
**    Copyright (C) 2011 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <mesh_io_off.hpp>

#include <string_helper.hpp>
#include <mesh.hpp>

#include <fstream>

namespace cpe
{
    mesh mesh_io_off::load_mesh_io_off(const std::string& filename)
    {

        std::ifstream stream(filename.c_str());

        if(stream.good()==false)
        {throw std::string("cannot open file in mesh_io_off::load_mesh_io_off("+filename+")");}

        std::string buffer;
        int N_vertex=0,N_polygon=0;//,N_edge=0;

        //toof flag
        bool is_toff=false;

        // look for *OFF*
        bool is_loop=true;
        while(stream.good() && is_loop==true)
        {
            std::getline(stream,buffer);
            if(buffer.length()>0
               && buffer[0]!='#'
               && buffer.find("OFF")!=std::string::npos)
            {
                is_loop=false;
                if(buffer.find("TOFF")!=std::string::npos)
                    is_toff=true;
            }
        }

        if(stream.good()==false && is_loop==true)
        {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") cannot find OFF header");}

        // look for N_vertex N_polygon N_edges
        is_loop=true;
        while(stream.good() && is_loop==true)
        {
            std::getline(stream,buffer);
            if(buffer.length()>0 && buffer[0]!='#')
            {
                std::vector <std::string> token=cpe::string_tokenizer::tokenize(buffer);
                if(token.size()>=3)
                {
                    bool vertex_converted=false;
                    bool polygon_converted=false;
                    N_vertex  = cpe::string_converter::value_of<int>(token[0],&vertex_converted);
                    N_polygon = cpe::string_converter::value_of<int>(token[1],&polygon_converted);
                    //N_edge    = cpe::string_converter::value_of<int>(token[2],&edge_converted);

                    if(vertex_converted==true && polygon_converted==true)
                        is_loop=false;
                }
            }
        }
        if(stream.good()==false && is_loop==true)
        {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") cannot read header");}


        if(N_vertex<=0 || N_polygon<=0)
        {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") error");}

        std::vector<v3> point_set;
        std::vector<double> texture;

        //now read N_vertex positions
        for(int k_vertex=0;k_vertex<N_vertex;++k_vertex)
        {
            if(stream.good()==false)
            {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") error EOF found");}

            std::getline(stream,buffer);
            if(buffer.size()>0 && buffer[0]!='#')
            {
                std::vector<std::string> v_token=cpe::string_converter::delete_empty(cpe::string_tokenizer::tokenize(buffer));
                if(v_token.size()<3)
                {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") bad vertex");}

                bool converted_x0=false,converted_x1=false,converted_x2=false;
                double x0=cpe::string_converter::value_of<double>(v_token[0],&converted_x0);
                double x1=cpe::string_converter::value_of<double>(v_token[1],&converted_x1);
                double x2=cpe::string_converter::value_of<double>(v_token[2],&converted_x2);

                if(is_toff)
                {
                    if(v_token.size()<5)
                    {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") bad texture");}

                    double t0=cpe::string_converter::value_of<double>(v_token[3]);
                    double t1=cpe::string_converter::value_of<double>(v_token[4]);
                    texture.push_back(t0); texture.push_back(t1);

                }

                if(converted_x0==false || converted_x1==false || converted_x2==false)
                {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") vertex conversion failed");}

                point_set.push_back(v3(x0,x1,x2));
            }
            else
                --k_vertex;
        }



        std::vector<unsigned int> connectivity;
        //now read N_polygons
        for(int k_polygon=0;k_polygon<N_polygon;++k_polygon)
        {
            std::getline(stream,buffer);
            if(buffer.size()>0 && buffer[0]!='#')
            {
                std::vector <std::string> v_token=cpe::string_tokenizer::tokenize(buffer);
                if(v_token.size()>=4)
                {
                    // get the size
                    bool poly_size_convert=false;
                    int poly_size=cpe::string_converter::value_of<int>(v_token[0],&poly_size_convert);
                    if(poly_size_convert==false)
                    {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") cannot load polygon");}

                    if(poly_size<3)
                    {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") bad polygon size");}

                    std::vector<unsigned int> temp_poly;
                    if(int(v_token.size()-1)<poly_size)
                    {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") token problem");}

                    int counter=0;
                    for(int k_current_polygon=0;k_current_polygon<poly_size;++k_current_polygon)
                    {
                        ++counter;

                        bool index_converted=false;
                        if(v_token[counter]!="")
                        {
                            temp_poly.push_back(cpe::string_converter::value_of<unsigned int>(v_token[counter],&index_converted));
                            if(index_converted==false)
                            {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") load polygon index");}
                        }
                        else
                            k_current_polygon--;
                    }

                    //triangulate
                    for(unsigned int k=0,N=temp_poly.size();k<N-2;++k)
                    {
                        connectivity.push_back(temp_poly[0]);
                        connectivity.push_back(temp_poly[k+1]);
                        connectivity.push_back(temp_poly[k+2]);
                    }

                }
                else
                {throw std::string("mesh_io_off::load_mesh_io_off("+filename+") something wrong in polygon load");}

            }
            else
                --k_polygon;

        }

        mesh m(std::make_pair(point_set,connectivity));
        if(is_toff)
            m.get_texture()=texture;



        return m;

    }


    void mesh_io_off::save_mesh_io_off(const std::string& filename,const mesh& m,const bool& is_toof)
    {

        bool is_texture=(m.get_texture().size()==2*m.vertex_number());

        std::ofstream stream(filename.c_str());
        if(stream.good()==false)
        {throw std::string("mesh_io_off::save_mesh_io_off("+filename+") cannot open file");}

        if(is_texture==true && is_toof==true)
            stream<<"T";
        stream<<"OFF"<<std::endl;

        stream<<m.vertex_number()<<" "<<m.triangle_number()<<" 0"<<std::endl;
        const std::vector<v3>& p_vertex=m.get_vertex();
        unsigned int N_vertex=m.vertex_number();
        for(unsigned int k_vertex=0;k_vertex<N_vertex;++k_vertex)
        {
            //write vertices
            stream<<p_vertex[k_vertex].x()<<" "<<p_vertex[k_vertex].y()<<" "<<p_vertex[k_vertex].z();

            //write texture if necessary (warning un-official support)
            if(is_texture==true && is_toof==true)
                stream<<" "<<m.get_texture_tx(k_vertex)<<" "<<m.get_texture_ty(k_vertex);

            stream<<std::endl;
        }

        //write connectivity
        unsigned int N_polygon=m.triangle_number();
        for(unsigned int k_poly=0;k_poly<N_polygon;++k_poly)
        {
            stream<<3<<" ";
            for(unsigned int k_dim=0;k_dim<3;++k_dim)
                stream<<m.get_connectivity()[3*k_poly+k_dim]<<" ";
            stream<<std::endl;
        }

        stream.close();
    }




}
