
/** TP 5ETI - CPE Lyon - 2011/2012 */

#include <scene.hpp>

#include <GL/glew.h>
#include <QString>
#include <QFont>
#include <glwidget.hpp>


#include <opengl_drawer.hpp>
#include <texture.hpp>
#include <mesh_io_off.hpp>
#include <string_helper.hpp>
#include <cmath>
#include <simplexnoise1234.hpp>

using namespace cpe;



std::vector<mesh> arbre1;
std::vector<std::vector<v3> > arbre1_n;

std::vector<std::vector<mesh> > foret;
std::vector<std::vector<std::vector<v3> > > foret_n;

mesh terrain;
std::vector<v3> terrain_n;
std::vector<v3> seed;

mesh sphere;
std::vector<v3> sphere_n;

mesh cree_cone(double radius,double hauteur,int N)
{
    mesh m;
    m.add_vertex(0,0,hauteur);

    m.get_texture().push_back(0);
    m.get_texture().push_back(0);

    for(int k=0;k<N;k++)
    {
        double u=double(k)/N;
        double x=radius*cos(u*2*M_PI);
        double y=radius*sin(u*2*M_PI);

        m.add_vertex(v3(x,y,0.0));

        m.get_texture().push_back(3*x);
        m.get_texture().push_back(3*y);


        if(k<N-1)
            m.add_triangle(k+1,k+2,0);
        else
            m.add_triangle(N,1,0);
    }

    m.add_vertex(0,0,0);
    m.get_texture().push_back(0);
    m.get_texture().push_back(0);

    for(int k=0;k<N;k++)
    {
        double u=double(k)/N;
        double x=radius*cos(u*2*M_PI);
        double y=radius*sin(u*2*M_PI);

        m.add_vertex(v3(x,y,0.0));

        m.get_texture().push_back(3*x);
        m.get_texture().push_back(3*y);

        if(k<N-1)
            m.add_triangle(N+1+k+1,N+1+k+2,N+1);
        else
            m.add_triangle(N+1+N,N+1+1,N+1);
    }

    return m;
}

mesh cree_tronc(double radius,double hauteur,int N)
{
    mesh m;
    for(int k=0;k<N;k++)
    {
        double u=double(k)/N;
        double x=radius*cos(u*2*M_PI);
        double y=radius*sin(u*2*M_PI);

        m.add_vertex(v3(x,y,0.0));
        m.add_vertex(v3(x,y,1.0*hauteur));

        m.get_texture().push_back(double(k)/(N-1));
        m.get_texture().push_back(0);
        m.get_texture().push_back(double(k)/(N-1));
        m.get_texture().push_back(2);

        if(k<N-1)
        {
            m.add_triangle(2*k,2*k+1,2*(k+1));
            m.add_triangle(2*k+1,2*(k+1)+1,2*(k+1));
        }
        else
        {
            m.add_triangle(2*k,2*k+1,0);
            m.add_triangle(2*k+1,1,0);
        }
    }
    return m;

//    return cree_cone(radius,radius,hauteur,N);
}

std::vector<mesh> cree_arbre(double hauteur,int N)
{
    double dim=0.45;
    double base=0.5;
    double offset=0.2;

    mesh tronc=cree_tronc(0.11*hauteur,base*hauteur,N);

    mesh leaf1=cree_cone(0.4*hauteur,dim*hauteur,N);
    mesh leaf2=leaf1;
    mesh leaf3=leaf1;



    leaf1.translate(v3(0,0,base*hauteur));
    leaf2.translate(v3(0,0,(base+offset)*hauteur));
    leaf3.translate(v3(0,0,(base+offset*2)*hauteur));


    std::vector<mesh> v_sphere;
    v_sphere.resize(3);



    for(int k_s=0;k_s<3;++k_s)
    {
        v_sphere[k_s]=sphere;

        int r_ball=rand()%3;
        int r_vertex=rand()%N;

        switch(r_ball)
        {
        case 0:
            v_sphere[k_s].translate(leaf1.get_vertex(r_vertex));
            break;
        case 1:
            v_sphere[k_s].translate(leaf2.get_vertex(r_vertex));
            break;
        case 2:
            v_sphere[k_s].translate(leaf3.get_vertex(r_vertex));
            break;

        }
    }

    mesh temp_sphere=sphere;
    temp_sphere.translate(v3(0,0,hauteur*(base+2*offset+dim)));
    v_sphere.push_back(temp_sphere);


//    mesh leaf2=cree_cone(0.38*hauteur,0.21*hauteur,0.20*hauteur,N);
//    leaf2.translate(v3(0,0,(0.45+0.25)*hauteur));

//    mesh leaf3=cree_cone(0.35*hauteur,0.0*hauteur,0.30*hauteur,N);
//    leaf3.translate(v3(0,0,(0.45+0.25+0.20)*hauteur));

    std::vector<mesh> output;
    output.push_back(tronc);
    output.push_back(leaf1);
    output.push_back(leaf2);
    output.push_back(leaf3);

    for(int k_s=0;k_s<v_sphere.size();++k_s)
        output.push_back(v_sphere[k_s]);

    for(int k=0;k<output.size();k++)
        output[k].scale(1.0/(base+2*offset+dim));

    return output;
}

double perlin_noise(const double& x,const double& y,const int& octave,const double& persitency)
{
    double temp=0.0;
    double frequency=1.0;
    double current_persistency=1.0;
    for(int k=0;k<octave;k++)
    {

        temp += current_persistency*(0.5+0.5*snoise2(x*frequency,y*frequency));
        frequency *= 2.0;
        current_persistency *= persitency;
    }

    //normalize
    if( fabs(current_persistency-1)>0.0001 )
        temp *= (1-persitency)/(1-current_persistency);

    return temp;
}

mesh cree_terrain(double size,double height,int N)
{
    mesh terrain;
    for(int kx=0;kx<N;++kx)
    {
        for(int ky=0;ky<N;++ky)
        {
            double ux=double(kx)/(N-1);
            double uy=double(ky)/(N-1);

            double h=perlin_noise(3*ux,3*uy,5,0.3);
            v3 p=v3(size*ux,size*uy,h*height);
            terrain.add_vertex(p);

            terrain.get_texture().push_back(3*ux);
            terrain.get_texture().push_back(3*uy);
        }
    }


    for(int kx=0;kx<N-1;++kx)
    {
        for(int ky=0;ky<N-1;++ky)
        {
            terrain.add_triangle(kx+N*ky,kx+1+N*ky,kx+1+N*(ky+1));
            terrain.add_triangle(kx+N*ky,kx+1+N*(ky+1),kx+N*(ky+1));
        }
    }

    terrain.translate(v3(-0.5*size,-0.5*size,0));
    return terrain;
}

double interp_terrain(double x,double y,double size_terrain,int N_terrain)
{
    double ux=double(N_terrain-1)/size_terrain*(x+size_terrain/2.0);
    double uy=double(N_terrain-1)/size_terrain*(y+size_terrain/2.0);


    int kx0=int(ux);
    int ky0=int(uy);

    double rx=ux-double(kx0);
    double ry=uy-double(ky0);

    int k_square=kx0+(N_terrain-1)*ky0;
    int k_triangle;
    if(rx>ry)
        k_triangle=2*k_square+1;
    else
        k_triangle=2*k_square;

//std::cout<<k_triangle<<std::endl;
    v3 x0=terrain.get_vertex(k_triangle,0);
    v3 x1=terrain.get_vertex(k_triangle,1);
    v3 x2=terrain.get_vertex(k_triangle,2);

    return 1.0/3.0*(x0.z()+x1.z()+x2.z());
}

scene::scene()
{
    is_debug_mode=false;

    is_debug_vertex_index=true;
    is_debug_face_index=true;
    is_debug_inverted_normal=false;
    is_debug_draw_vertex=true;
}

void scene::load_model()
{

    sphere=mesh::load_off_file("data/sphere.off");
    sphere.scale(0.035);
    sphere_n=sphere.normal_vertex();

    arbre1=cree_arbre(1.0,18);
    for(int k=0;k<arbre1.size();++k)
        arbre1_n.push_back(arbre1[k].normal_vertex());


    //opengl_drawer::generate_texture(1,"data/seemless_grass_texture.ppm");
    opengl_drawer::generate_texture(1,"data/terrain.ppm");
    opengl_drawer::generate_texture(2,"data/trunc_seamless.ppm");
    opengl_drawer::generate_texture(3,"data/sapin.ppm");


    double size_terrain=20;
    int N_terrain=8;




    terrain=cree_terrain(size_terrain,1.0,N_terrain);

    int N_arbre=2;
    foret.resize(N_arbre);
    foret_n.resize(N_arbre);

    for(int k=0;k<N_arbre;++k)
    {
        std::cout<<k<<"/"<<N_arbre<<std::endl;

        double tx=size_terrain*(0.5-(rand()%50000)/50000.0);
        double ty=size_terrain*(0.5-(rand()%50000)/50000.0);

        double h=interp_terrain(tx,ty,size_terrain,N_terrain);

        double r=0.5-double(rand()%50000)/50000.0;
        foret[k]=cree_arbre(1.0+0.5*r,18);

        for(int kk=0;kk<foret[k].size();++kk)
        {
            foret[k][kk].translate(v3(tx,ty,h));
            foret_n[k].push_back(foret[k][kk].normal_vertex());
        }

        bool is_good=true;
        for(int kk=0;is_good==true && kk<seed.size();++kk)
        {
            if( (seed[kk]-v3(tx,ty,0.0)).norm() < 1)
            {
                k--;
                is_good=false;
            }
        }
        if(is_good==true)
            seed.push_back(v3(tx,ty,0.0));
    }

//    double h=interp_terrain(0.0,0.0,size_terrain,N_terrain);
//    for(int k=0;k<arbre1.size();k++)
//        arbre1[k].translate(v3(0,0,h-0.05));


    terrain_n=terrain.normal_vertex();
    counter=0;
}

void scene::draw_scene()
{
    //affichage de la scène

    glEnable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);


    //affiche le maillage 0
    glColor3d(0.9,0.9,0.9);
    glBindTexture(GL_TEXTURE_2D,2);
    //cpe::opengl_drawer::draw(mesh_0,mesh_0_normal);

    //affiche le maillage 1
    glColor3d(0.6,0.9,0.6);
    //cpe::opengl_drawer::draw(mesh_1,mesh_1_normal);


    //glColor3d(0.6,0.6,0.4);


//    glDisable(GL_TEXTURE_2D);
//    //mesh m=cree_tronc(1.0/5.0,1.0,60);

//    //mesh m=cree_cone(1.0,1.0,9);
//    mesh m=cree_arbre(1.0,29)[1];
//    mesh m1=cree_arbre(1.0,29)[2];
//    mesh m2=cree_arbre(1.0,29)[3];


//    glEnable(GL_LIGHTING);

//    //glColor3d(0.75,0.75,0.75);
//    glColor3d(0.75,0.95,0.75);

//    std::vector<v3> nn=m.normal_vertex();
//    std::vector<v3> nn1=m1.normal_vertex();
//    std::vector<v3> nn2=m2.normal_vertex();

//    glEnable(GL_POLYGON_OFFSET_FILL);
//    glPolygonOffset(1,1);
//    opengl_drawer::draw(m,nn);
//    opengl_drawer::draw(m1,nn1);
//    opengl_drawer::draw(m2,nn2);
//    glDisable(GL_POLYGON_OFFSET_FILL);



//    glDisable(GL_LIGHTING);

//    glColor3d(0,0,0);
//    glLineWidth(3.0);
////    opengl_drawer::draw(m);
////    opengl_drawer::draw(m1);
////    opengl_drawer::draw(m2);

//        draw_debug(m);


//    for(int k=0;k<arbre1.size();k++)
//    {
//        if(k==0)
//            glColor3d(0.20,0.12,0.12);
//        else if (k<3)
//            glColor3d(0.2,0.35,0.2);
//        else
//        {
//            glColor3d(1,0,0);
//        }
//        opengl_drawer::draw(arbre1[k],arbre1_n[k]);
//    }



    for(int k=0;k<foret.size();++k)
    {
        for(int k2=0;k2<foret[k].size();++k2)
        {
            if(k2==0)
            {
                //glColor3d(0.20,0.12,0.12);
                //glColor3d(1,1,1);
                glColor3d(0.20*3,0.12*3,0.12*3);
                glEnable(GL_TEXTURE_2D);
                glBindTexture(GL_TEXTURE_2D,2);
            }
            else if (k2<4)
            {
                glColor3d(0.2*2,0.35*2,0.2*2);
                glEnable(GL_TEXTURE_2D);
                glBindTexture(GL_TEXTURE_2D,3);
            }
            else if (k2==7)
            {
                glDisable(GL_TEXTURE_2D);
                glColor3d(0.9,0.9,0.1);
            }
            else
            {
                glDisable(GL_TEXTURE_2D);
                int r=int(3*(1+cos(5000.0*k*k2+500*k))/2.0);
                if(r==0)
                    glColor3d(1,0,0);
                if(r==1)
                    glColor3d(0,1,0);
                if(r==2)
                    glColor3d(0,0,1);
            }
            glDisable(GL_TEXTURE_2D);
            opengl_drawer::draw(foret[k][k2],foret_n[k][k2]);
        }
    }


    glEnable(GL_POLYGON_OFFSET_FILL);
       glPolygonOffset(1,1);


//    glDisable(GL_TEXTURE_2D);
//    glColor3d(0.7,0.8,0.7);
//    glColor3d(0.9,0.9,0.9);
//    glEnable(GL_TEXTURE_2D);
//    glBindTexture(GL_TEXTURE_2D,1);
//    opengl_drawer::draw(terrain,terrain_n);

        glDisable(GL_POLYGON_OFFSET_FILL);

         glColor3d(0,0,0);
         glLineWidth(1.5);
opengl_drawer::draw(terrain);




//    double t=double(counter)/1000.0;
//    glColor3d(0,0,0);
//    v3 x0(1.5+3*cos(t),4.2,-5);
//    v3 x1(1.5+3*cos(t),4.2,+5);
//    double h=interp_terrain(1.5+3*cos(t),4.2,20,10);
//   //exit(0);
//    glBegin(GL_LINES);
//    glVertex3dv(&x0.x());
//    glVertex3dv(&x1.x());
//    glEnd();

//    draw_debug(terrain);
    ++counter;


}

void scene::creation_damier()
{
    opengl_drawer::generate_texture(1,"data/damier.ppm");
    mesh_plan=mesh::load_off_file("data/quad.off");
    mesh_plan.scale_unity();
    mesh_plan.scale(2.0);
    mesh_plan_normal=mesh_plan.normal_vertex();
}

void scene::affichage_damier()
{
    //affiche le plan avec la texture
    glBindTexture(GL_TEXTURE_2D,1);
    glColor3d(0.9,0.9,0.9);
    opengl_drawer::draw(mesh_plan,mesh_plan_normal);

    //affiche les axes
    glDisable(GL_LIGHTING);
    glBegin(GL_LINES);
    glColor3d(1,0,0);
    glVertex3d(0,0,0);
    glVertex3d(1.25,0,0);

    glColor3d(0,1,0);
    glVertex3d(0,0,0);
    glVertex3d(0,1.25,0);

    glColor3d(0,0,1);
    glVertex3d(0,0,0);
    glVertex3d(0,0,1.25);
    glEnd();
}

void scene::draw_debug(cpe::mesh mesh_current)
{

    if(is_debug_mode==false)
        return ;

    double L_text=0.04;
    if(is_debug_inverted_normal==true)
        L_text*=-1;


    std::vector<v3> nv=mesh_current.normal_vertex();
    std::vector<v3> np=mesh_current.normal_polygon();

    glEnable(GL_LIGHTING);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(5,5);
//    if(is_plain==true)
//        MC_opengl_drawer::draw(mesh);


    glColor3d(0.5,0.1,0.0);

    glDisable(GL_POLYGON_OFFSET_FILL);
    glDisable(GL_LIGHTING);



//    glColor3d(0,0,1);
//    if(qglwidget_helper.get_is_grid()==true)
//        MC_opengl_drawer::draw_grid(mesh);

    glColor3d(0.2,0.2,1);
    glPointSize(9.0);
    if(is_debug_draw_vertex==true)
    {
        glBegin(GL_POINTS);
        for(int k=0,N=mesh_current.vertex_number();k<N;++k)
        {
            const v3& p=mesh_current.get_vertex(k);
            glVertex3d(p.x(),p.y(),p.z());
        }
        glEnd();
    }

    std::vector<v3> p_index;
    std::vector<v3> translation_face;
    for(int k=0,N=mesh_current.vertex_number();k<N;++k)
        p_index.push_back(mesh_current.get_vertex(k)-nv[k]*L_text);
    for(int k=0,N=mesh_current.triangle_number();k<N;++k)
        translation_face.push_back(-L_text*np[k]);

    glColor3d(1,0,0);
    draw_index(p_index);
    glColor3d(1,0,0);
    draw_index_face(mesh_current,translation_face);
}

void scene::draw_index(const std::vector<cpe::v3>& x)
{
    if(is_debug_vertex_index==false)
        return ;

    int N=x.size();
    for(int k=0;k<N;++k)
        p_widget->draw_text(x[k][0],x[k][1],x[k][2]-0.05,QString(cpe::string_converter::to_string(k).c_str()),QFont("Times",18,QFont::Normal));
}

void scene::draw_index_face(const cpe::mesh& mesh,const std::vector<v3>& translate)
{
    if(is_debug_face_index==false)
        return ;

    cpe::v3 x_bar;
    int N=mesh.triangle_number();
    for(int k=0;k<N;++k)
    {
        x_bar=(mesh.get_vertex(k,0)+mesh.get_vertex(k,1)+mesh.get_vertex(k,2))/3.0+translate[k];
        p_widget->draw_text(x_bar[0],x_bar[1],x_bar[2],QString(string_converter::to_string(k).c_str()),QFont("Times",18,QFont::Normal));
    }
}

void scene::set_glwidget(glwidget *_p_widget)
{
    p_widget=_p_widget;
}

void scene::set_is_debug_mode(bool value)
{
    is_debug_mode=value;
}

void scene::set_is_debug_vertex_index(bool value)
{
    is_debug_vertex_index=value;
}

void scene::set_is_debug_face_index(bool value)
{
    is_debug_face_index=value;
}

void scene::set_is_debug_inverted_normal(bool value)
{
    is_debug_inverted_normal=value;
}
void scene::set_is_debug_vertex_point(bool value)
{
    is_debug_draw_vertex=value;
}
